package abchr;

import abchr.audio.Sample;

import java.io.IOException;
import java.io.Serializable;
import java.util.*;
import java.text.DateFormat;
import java.text.SimpleDateFormat;

import org.jdom.Element;
import org.jdom.Document;

public class ABXModule extends AbstractProjectModule {
	private static class MapKey implements Serializable {
		private final Sample aSample,bSample;

		public MapKey(Sample aIndex,Sample bIndex) {
			this.aSample=aIndex;
			this.bSample=bIndex;
		}

		public boolean equals(Object o) {
			if(this==o) return true;
			if(!(o instanceof MapKey)) return false;
			final MapKey mapKey=(MapKey)o;
			return mapKey.aSample==aSample && mapKey.bSample==bSample;
		}

		public int hashCode(){return 29*aSample.hashCode()+bSample.hashCode();}
	}

	// maps (index of A,index of B) -> ABX test
	// provides fast access to tests
	private transient Map<MapKey,ABXTest> testMap;

	private boolean[][] finished;
	private double pValThreshold=0;

	private SampleConfigModule sampleConfig;
	private ComparisonModule compModule;

	public ABXModule(SampleConfigModule sampleConfig,ComparisonModule compModule) {
		this.sampleConfig=sampleConfig;
		this.compModule=compModule;
		int n=sampleConfig.getSampleCount()+1;
		testMap=new HashMap<MapKey,ABXTest>(2*n);
		finished=new boolean[n][n];
	}

	// tests are created lazily since most of the
	// sampleCount^2 tests will probably never be used
	public ABXTest getTest(Sample aSample,Sample bSample) {
		if(aSample==bSample){throw new IllegalArgumentException("Indices equal or out of range");}
		if(sampleConfig.indexOf(aSample)>sampleConfig.indexOf(bSample)) {
			Sample t=aSample;
			aSample=bSample;
			bSample=t;
		}
		MapKey key=new MapKey(aSample,bSample);
		if(testMap.containsKey(key)){return (ABXTest)testMap.get(key);}
		ABXTest abxTest=new ABXTestImpl(aSample,bSample);
		testMap.put(key,abxTest);
		return abxTest;
	}

	public void resetTest(Sample aSample,Sample bSample) {
		if(aSample==bSample){throw new IllegalArgumentException("Indices equal or out of range");}
		MapKey key=new MapKey(aSample,bSample);
		if(!testMap.containsKey(key)){return;}
		ABXTest abxTest=new ABXTestImpl(aSample,bSample);
		testMap.put(key,abxTest);
	}

	public SampleConfigModule getSampleConfig(){return sampleConfig;}

	public boolean isFinished(int aSample,int bSample) {
		return finished[aSample][bSample];
	}

	public void setFinished(int aSample,int bSample,boolean finished) {
		this.finished[aSample][bSample]=finished;
		if(finished && aSample==0 && getTest(sampleConfig.getOriginalSample(),sampleConfig.getSample(bSample-1)).pVal()<pValThreshold) {
			compModule.getGroup(bSample-1).setBlind(false);
		}
	}

	public double getPValThreshold(){return pValThreshold;}
	public void setPValThreshold(double pValThreshold){this.pValThreshold=pValThreshold;}

	public Element toResultsXML() {
		Element root=new Element("ABX");
		Element e;
		DateFormat dateFormat=new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ssZ",Locale.getDefault());
		DetailedABXTestFormatter.SimpleRangeFormatter rangeFormatter=new DetailedABXTestFormatter.SimpleRangeFormatter();
		for(Map.Entry<MapKey,ABXTest> test : testMap.entrySet()) {
			if(test.getValue().getNumTrials()==0){continue;}
			e=new Element("Comparison");
			e.addContent(new Element("SampleA").setText(Integer.toString(sampleConfig.indexOf(test.getKey().aSample))));
			e.addContent(new Element("SampleB").setText(Integer.toString(sampleConfig.indexOf(test.getKey().bSample))));
			e.addContent(new Element("TotalTrials").setText(Integer.toString(test.getValue().getNumTrials())));
			e.addContent(new Element("CorrectTrials").setText(Integer.toString(test.getValue().getCorrect())));
			Element progress=new Element("Progress");
			Element trialNode;
			ABXTrial trial;
			for(int i=0;i<test.getValue().getNumTrials();i++) {
				trial=test.getValue().getTrial(i);
				trialNode=new Element("Trial");
				trialNode.addContent(new Element("Correct").setText(trial.getCorrect()?"true":"false"));
				trialNode.addContent(new Element("Time").setText(dateFormat.format(new Date(trial.getTimeStamp()))));
				trialNode.addContent(new Element("RangeStart").setText(rangeFormatter.formatTime(trial.getRangeBegin())));
				trialNode.addContent(new Element("RangeEnd").setText(rangeFormatter.formatTime(trial.getRangeEnd())));
				progress.addContent(trialNode);
			}
			e.addContent(progress);
			root.addContent(e);
		}
		return root;
	}

	private void writeObject(java.io.ObjectOutputStream out) throws IOException {
		//System.out.println("Writing ABXModule");
		out.defaultWriteObject();
		Set<Map.Entry<MapKey,ABXTest>> tests=testMap.entrySet();
		for(Iterator<Map.Entry<MapKey,ABXTest>> it=tests.iterator();it.hasNext();) {
			if(it.next().getValue().getNumTrials()==0){it.remove();}
		}
		out.writeInt(tests.size());
		for(Map.Entry<MapKey,ABXTest> test : tests) {
			out.writeObject(test.getKey());
			out.writeObject(test.getValue());
		}
		//System.out.println("Done writing ABXModule");
	}

	private void readObject(java.io.ObjectInputStream in) throws IOException, ClassNotFoundException {
		in.defaultReadObject();
		int numTests=in.readInt();
		ABXTest test;
		MapKey key;
		testMap=new HashMap<MapKey,ABXTest>(2*sampleConfig.getSampleCount());
		for(int i=0;i<numTests;i++) {
			key=(MapKey)in.readObject();
			test=(ABXTest)in.readObject();
			testMap.put(key,test);
		}
	}

	public static ProjectModule createInstance(Document doc,SampleConfigModule sampleConfig,ComparisonModule compModule) {
		ABXModule m=new ABXModule(sampleConfig,compModule);
		Element e=doc.getRootElement().getChild("abx");
		if(e!=null) {
			String s=e.getChildText("revealrefpval");
			if(s!=null) {
				m.setPValThreshold(Double.parseDouble(s));
			}
		}
		return m;
	}
}