package abchr;

import org.jdom.Document;
import org.jdom.Element;

import javax.xml.parsers.FactoryConfigurationError;
import java.io.*;
import java.util.Iterator;
import java.util.List;

public class IniConfigIO implements ConfigIO {
	private static interface Converter {
		public Element toXML(IniStringMap iniMap);
		public void writeToIni(BufferedWriter writer,org.jdom.Document doc) throws IOException;
	}

	private static class TestNameConverter implements Converter {
		public Element toXML(IniStringMap iniMap) {
			return new Element("testname").setText(iniMap.get("TestName"));
		}

		public void writeToIni(BufferedWriter writer,Document doc) throws IOException {
			writer.write("TestName = "+doc.getRootElement().getChild("testname").getText());
			writer.newLine();
		}
	}

	private static class TestConfigConverter implements Converter {
		public Element toXML(IniStringMap iniMap) {
			Element root=new Element("testconfig");
			Element sample=new Element("original");
			sample.addContent(new Element("location").setText(iniMap.get("Original")));
			sample.addContent(new Element("offset").setText(iniMap.contains("Offset_Orig")?iniMap.get("Offset_Orig"):"0"));
			sample.addContent(new Element("gain").setText(iniMap.contains("Gain_Orig")?iniMap.get("Gain_Orig"):"0"));
			root.addContent(sample);
			int i=1;
			while(iniMap.contains("Sample"+i)) {
				sample=new Element("testsample");
				sample.addContent(new Element("location").setText(iniMap.get("Sample"+i)));
				sample.addContent(new Element("offset").setText(iniMap.contains("Offset"+i)?iniMap.get("Offset"+i):"0"));
				sample.addContent(new Element("gain").setText(iniMap.contains("Gain"+i)?iniMap.get("Gain"+i):"0"));
				root.addContent(sample);
				i++;
			}
			return root;
		}

		public void writeToIni(BufferedWriter writer,Document doc) throws IOException {
			Element root=doc.getRootElement().getChild("testconfig");
			Element sample=root.getChild("original");
			writer.write("Original = ");
			writer.write(sample.getChild("location").getText());
			writer.newLine();
			List samples=root.getChildren("testsample");
			int i=1;
			for(Iterator it=samples.iterator();it.hasNext();) {
				sample=(org.jdom.Element)it.next();
				writer.write("Sample"+(i++)+" = ");
				writer.write(sample.getChild("location").getText());
				writer.newLine();
			}
			writer.newLine();
			sample=root.getChild("original");
			writer.write("Offset_Orig = ");
			writer.write(sample.getChild("offset").getText());
			writer.newLine();
			i=1;
			for(Iterator it=samples.iterator();it.hasNext();) {
				sample=(org.jdom.Element)it.next();
				writer.write("Offset"+(i++)+" = ");
				writer.write(sample.getChild("offset").getText());
				writer.newLine();
			}
			writer.newLine();
			sample=root.getChild("original");
			writer.write("Gain_Orig = ");
			writer.write(sample.getChild("gain").getText());
			writer.newLine();
			i=1;
			for(Iterator it=samples.iterator();it.hasNext();) {
				sample=(org.jdom.Element)it.next();
				writer.write("Gain"+(i++)+" = ");
				writer.write(sample.getChild("gain").getText());
				writer.newLine();
			}
		}
	}

	private static class RatingScaleConverter implements Converter {
		public Element toXML(IniStringMap iniMap) {
			if(!iniMap.contains("ScaleLabel5")){return null;}
			Element root=new Element("ratingscale");
			Element label;
			for(int i=5;i>=1;i--) {
				label=new Element("label");
				root.addContent(iniMap.contains("ScaleLabel"+i)?label.setText(iniMap.get("ScaleLabel"+i)):label);
			}
			return root;
		}

		public void writeToIni(BufferedWriter writer,org.jdom.Document doc) throws IOException {
			Element root=doc.getRootElement().getChild("ratingscale");
			if(root==null){return;}
			List children=root.getChildren("label");
			String label;
			int i=5;
			for(Iterator it=children.iterator();it.hasNext();) {
				label=((Element)it.next()).getText();
				if(!label.equals("")){writer.write("ScaleLabel"+i+" = "+label);writer.newLine();}
				i--;
				if(i<1){break;}
			}
		}
	}

	private Converter[] converters=new Converter[]{new TestNameConverter(),new TestConfigConverter(),new RatingScaleConverter()};

	public Document readConfig(InputStream in) throws IOException {
		try {
			IniStringMap iniMap=new IniStringMap(in);
			org.jdom.Element root=new org.jdom.Element("config");
			for(int i=0;i<converters.length;i++) {
				org.jdom.Element element=converters[i].toXML(iniMap);
				if(element!=null){root.addContent(element);}
			}
			return new org.jdom.Document(root);
		} catch(FactoryConfigurationError factoryConfigurationError) {
			throw new IOException("Error initializing XML input");
		} catch(ParseException e) {
			throw new IOException("Wrong input format");
		}
	}

	public void writeConfig(OutputStream out,Document doc) throws IOException {
		//System.out.println("in IniConfigIO");
		BufferedWriter writer=new BufferedWriter(new OutputStreamWriter(out));
		for(int i=0;i<converters.length;i++) {
			converters[i].writeToIni(writer,doc);
			writer.newLine();
		}
		writer.flush();
	}
}
