package guiutils;

import javax.swing.*;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import java.awt.*;
import java.awt.event.*;
import java.util.ArrayList;
import java.util.List;

public class RangeSelectBar extends JComponent implements MouseListener, MouseMotionListener, ChangeListener {
	private BoundedRangeModel model;

	private SavedRangesModel savedRanges=new SavedRangesModel();

	private final List listeners=new ArrayList(2);

	public RangeSelectBar(BoundedRangeModel model) {
		this.model=model;
		model.addChangeListener(this);
		savedRanges.addListener(this);
		this.setPreferredSize(new Dimension(300,20));
		this.addMouseListener(this);
		this.addMouseMotionListener(this);

		InputMap im=this.getInputMap(JComponent.WHEN_IN_FOCUSED_WINDOW);
		ActionMap am=this.getActionMap();
		class RangeSaveAction extends AbstractAction {
			int index;
			public RangeSaveAction(int index){this.index=index;}
			public void actionPerformed(ActionEvent e) {
				savedRanges.set(index,new Range(RangeSelectBar.this.model.getValue(),RangeSelectBar.this.model.getExtent()));
			}
		}
		class RangeLoadAction extends AbstractAction {
			int index;
			public RangeLoadAction(int index){this.index=index;}
			public void actionPerformed(ActionEvent e) {
				if(savedRanges.get(index)==null){return;}
				Range r=savedRanges.get(index);
				BoundedRangeModel model=RangeSelectBar.this.model;
				model.setRangeProperties(r.getBegin(),r.getExtent(),model.getMinimum(),model.getMaximum(),false);
			}
		}
		for(int i=1;i<=9;i++) {
			im.put(KeyStroke.getKeyStroke(Character.forDigit(i,10),InputEvent.ALT_MASK),"saveRange"+i);
			am.put("saveRange"+i,new RangeSaveAction(i-1));
			im.put(KeyStroke.getKeyStroke(Character.forDigit(i,10),InputEvent.CTRL_MASK),"loadRange"+i);
			am.put("loadRange"+i,new RangeLoadAction(i-1));
		}
	}

	// start, end in millseconds
	public RangeSelectBar(int start,int end) {
		this(new DefaultBoundedRangeModel(start,end-start,start,end));
	}

	public void stateChanged(ChangeEvent e) {
		notifyListeners();
		repaint();
	}

	public BoundedRangeModel getModel(){return model;}
	public void setModel(BoundedRangeModel model) {
		model.removeChangeListener(this);
		this.model=model;
		model.addChangeListener(this);
		notifyListeners();
		repaint();
	}

	public SavedRangesModel getSavedRanges(){return savedRanges;}
	public void setSavedRanges(SavedRangesModel model) {
		savedRanges.removeListener(this);
		savedRanges=model;
		savedRanges.addListener(this);
		repaint();
	}

	public void addListener(RangeSelectBarListener listener){listeners.add(listener);}
	public void removeListener(RangeSelectBarListener listener){listeners.remove(listener);}

	private void notifyListeners() {
		for(int i=0;i<listeners.size();i++) {
			((RangeSelectBarListener)listeners.get(i)).selectionChanged(this,model.getValue(),model.getValue()+model.getExtent());
		}
	}

	private static Color[] colors={Color.GRAY,Color.ORANGE,Color.WHITE,Color.CYAN};

	protected void paintComponent(Graphics g) {
		super.paintComponent(g);
		g.setColor(Color.BLACK);
		g.drawRect(0,0,this.getWidth()-1,this.getHeight()-1);
		if(this.isEnabled()) {
			int xStart=rangeToDisplay(model.getValue());
			int xEnd=rangeToDisplay(model.getValue()+model.getExtent());
			g.setColor(UIManager.getColor("ProgressBar.foreground"));
			g.fillRect(xStart,2,xEnd-xStart,this.getHeight()-5);
			g.setColor(Color.BLACK);
			g.drawRect(xStart,2,xEnd-xStart,this.getHeight()-5);
			g.setColor(Color.BLACK);
			Range r;
			for(int i=0;i<9;i++) {
				r=savedRanges.get(i);
				if(r!=null) {
					xStart=rangeToDisplay(r.getBegin());
					xEnd=rangeToDisplay(r.getBegin()+r.getExtent());
					g.setColor(colors[i%4]);
					g.drawRect(xStart,4,xEnd-xStart,this.getHeight()-9);
					g.drawString(Integer.toString(i+1),xStart+(xEnd-xStart)/2,this.getHeight()/2+5);
				}
			}
		}
	}

	// value in milliseconds
	protected int rangeToDisplay(int value) {
		return 2+((this.getWidth()-5)*(value-model.getMinimum()))/(model.getMaximum()-model.getMinimum());
	}

	protected int displayToRange(int pos) {
		int r=model.getMinimum()+((model.getMaximum()-model.getMinimum())*(pos-2))/(this.getWidth()-5);
		return r<model.getMinimum()?model.getMinimum():r>model.getMaximum()?model.getMaximum():r;
	}

	private int startX;

	private static final int NEW_RANGE=0,LOWER_BOUNDARY=1,UPPER_BOUNDARY=2,MOVE_RANGE=3;
	private int action=-1;

	private int oldStart,oldEnd;

	public void mouseClicked(MouseEvent e) {
		if(SwingUtilities.isLeftMouseButton(e) && e.getClickCount()==2) {
			model.setRangeProperties(model.getMinimum(),model.getMaximum()-model.getMinimum(),model.getMinimum(),model.getMaximum(),false);
		}
	}

	public void mouseEntered(MouseEvent e){}
	public void mouseExited(MouseEvent e){}

	public void mousePressed(MouseEvent e) {
		if(!isEnabled()){return;}
		oldStart=model.getValue();
		oldEnd=model.getValue()+model.getExtent();
		model.setValueIsAdjusting(true);
		if(Math.abs(rangeToDisplay(model.getValue())-e.getX())<3) {
			action=LOWER_BOUNDARY;
		} else if(Math.abs(rangeToDisplay(model.getValue()+model.getExtent())-e.getX())<3) {
			action=UPPER_BOUNDARY;
		} else if(e.getX()>rangeToDisplay(model.getValue()) && e.getX()<rangeToDisplay(model.getValue()+model.getExtent()) && SwingUtilities.isRightMouseButton(e)) {
			action=MOVE_RANGE;
			startX=e.getX();
		} else {
			action=NEW_RANGE;
			startX=e.getX();
		}
	}

	public void mouseReleased(MouseEvent e) {
		if(!isEnabled()){return;}
		startX=-1;
		action=-1;
		mouseDragged(e);
		model.setValueIsAdjusting(false);
	}

	public void mouseDragged(MouseEvent e) {
		if(action!=-1) {
			int a=0,b=0;
			switch(action) {
			case NEW_RANGE:
				a=displayToRange(startX);
				b=displayToRange(e.getX());
				break;
			case LOWER_BOUNDARY:
				a=displayToRange(e.getX());
				b=oldEnd;
				break;
			case UPPER_BOUNDARY:
				a=oldStart;
				b=displayToRange(e.getX());
				break;
			case MOVE_RANGE:
				int offset=displayToRange(e.getX())-displayToRange(startX);
				a=oldStart+offset;
				b=oldEnd+offset;
			}
			if(b<a){int temp=a;a=b;b=temp;}
			a=a<model.getMinimum()?model.getMinimum():a;
			b=b>model.getMaximum()?model.getMaximum():b;
			model.setRangeProperties(a,b-a,model.getMinimum(),model.getMaximum(),true);
			repaint();
		}
	}

	public void mouseMoved(MouseEvent e) {
		if(isEnabled() && (Math.abs(rangeToDisplay(model.getValue())-e.getX())<3 || Math.abs(rangeToDisplay(model.getValue()+model.getExtent())-e.getX())<3)) {
			setCursor(Cursor.getPredefinedCursor(Cursor.W_RESIZE_CURSOR));
		} else {
			setCursor(Cursor.getDefaultCursor());
		}
	}
}